from arkitekt.model import Manifest
from .types import EasyApp
from arkitekt.apps.fallbacks import ImportException
from arkitekt.apps.service.fakts import build_arkitekt_fakts, build_arkitekt_token_fakts
from arkitekt.apps.service.herre import build_arkitekt_herre
from typing import Optional, List


def build_arkitekt_easy_app(
    manifest: Manifest,
    url: Optional[str] = None,
    no_cache: bool = False,
    headless: bool = False,
    instance_id: Optional[str] = None,
    token: Optional[str] = None,
    enforce: Optional[List[str]] = None,
):
    fakts = (
        build_arkitekt_fakts(
            manifest=manifest, url=url, no_cache=no_cache, headless=headless
        )
        if not token
        else build_arkitekt_token_fakts(
            manifest=manifest,
            token=token,
            url=url,
            no_cache=no_cache,
            headless=headless,
        )
    )
    herre = build_arkitekt_herre(fakts=fakts)

    try:
        from arkitekt.apps.service.rekuest import build_arkitekt_rekuest

        rekuest = build_arkitekt_rekuest(
            fakts=fakts, herre=herre, instance_id=instance_id or "main"
        )
    except ImportError as e:
        if enforce and "rekuest" in enforce:
            raise e
        rekuest = ImportException(import_exception=e, install_library="rekuest")

    try:
        from arkitekt.apps.service.mikro import build_arkitekt_mikro

        mikro = build_arkitekt_mikro(fakts=fakts, herre=herre)
    except ImportError as e:
        if enforce and "mikro" in enforce:
            raise e
        mikro = ImportException(import_exception=e, install_library="mikro")

    try:
        from arkitekt.apps.service.unlok import build_arkitekt_unlok

        unlok = build_arkitekt_unlok(herre=herre, fakts=fakts)
    except ImportError as e:
        if enforce and "unlok" in enforce:
            raise e
        unlok = ImportException(import_exception=e, install_library="unlok")

    try:
        from arkitekt.apps.service.fluss import build_arkitekt_fluss

        fluss = build_arkitekt_fluss(herre=herre, fakts=fakts)
    except ImportError as e:
        if enforce and "fluss" in enforce:
            raise e
        fluss = ImportException(import_exception=e, install_library="fluss")

    try:
        from arkitekt.apps.service.omero_ark import build_arkitekt_omero_ark

        omero_ark = build_arkitekt_omero_ark(herre=herre, fakts=fakts)
    except ImportError as e:
        if enforce and "omero_ark" in enforce:
            raise e
        omero_ark = ImportException(import_exception=e, install_library="omero_ark")

    return EasyApp(
        manifest=manifest,
        fakts=fakts,
        herre=herre,
        rekuest=rekuest,
        mikro=mikro,
        unlok=unlok,
        fluss=fluss,
        omero_ark=omero_ark,
    )



import rich_click as click
from arkitekt.cli.options import *
import asyncio
from arkitekt.cli.ui import construct_run_panel
from importlib import import_module
from .utils import import_builder


async def run_app(app):
    async with app:
        await app.rekuest.run()


@click.command("prod")
@click.option(
    "--url",
    help="The fakts url for connection",
    default="http://localhost:8000",
    envvar="FAKTS_URL",
)
@with_builder
@with_token
@with_instance_id
@with_headless
@with_log_level
@with_version
@with_skip_cache
@click.pass_context
def prod(ctx, version=None, url=None, entrypoint=None, builder=None, **builder_kwargs):
    """Runs the app in production mode

    \n
    You can specify the builder to use with the --builder flag. By default, the easy builder is used, which is designed to be easy to use and to get started with.

    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)
    entrypoint = entrypoint or manifest.entrypoint

    builder = import_builder(builder)

    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e

    app = builder(
        **manifest.to_builder_dict(),
        **builder_kwargs,
    )

    panel = construct_run_panel(app)
    console.print(panel)

    asyncio.run(run_app(app))



from pydantic import BaseModel, Field, validator
import datetime
from typing import List, Optional
from enum import Enum
import semver
import uuid
from rekuest.api.schema import DefinitionInput


class Framework(str, Enum):
    """Do we support other frameworks?"""

    VANILLA = "vanilla"
    TENSORFLOW = "tensorflow"
    PYTORCH = "pytorch"


class Requirement(str, Enum):
    """ """

    GPU = "gpu"


class Packager(str, Enum):
    CONDA = "conda"
    POETRY = "poetry"
    PIP = "pip"


class Manifest(BaseModel):
    identifier: str
    version: str
    author: str
    logo: Optional[str]
    entrypoint: str
    scopes: List[str]
    requirements: List[Requirement] = Field(default_factory=list)
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)

    @validator("version", pre=True)
    def version_must_be_semver(cls, v) -> str:
        """Checks that the version is a valid semver version"""
        if isinstance(v, str):
            try:
                semver.VersionInfo.parse(v)
            except ValueError:
                raise ValueError("Version must be a valid semver version")
        return str(v)

    def to_console_string(self):
        return f"📦 {self.identifier} ({self.version}) by {self.author}"

    def to_builder_dict(self):
        return {
            "identifier": self.identifier,
            "version": self.version,
            "logo": self.logo,
            "scopes": self.scopes,
        }

    class Config:
        validate_assignment = True


class Deployment(BaseModel):
    """A deployment is a Release of a Build.
    It contains the build_id, the manifest, the builder, the definitions, the image and the deployed_at timestamp.



    """

    deployment_id: str = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="The unique identifier of the deployment",
    )
    manifest: Manifest = Field(description="The manifest of the app that was deployed")
    builder: str = Field(
        description="The builder that was used to build the app. CUrrently always port"
    )
    build_id: str = Field(
        description="The build_id of the build that was deployed. Is referenced in the build.yaml file."
    )
    definitions: List[DefinitionInput] = Field(
        description="Definitions of nodes that are contained in the app."
    )
    image: str = Field(
        description="The docker image that was built for this deployment"
    )
    deployed_at: datetime.datetime = Field(
        default_factory=datetime.datetime.now,
        description="The timestamp of the deployment",
    )


class DeploymentsConfigFile(BaseModel):
    """The ConfigFile is a pydantic model that represents the deployments.yaml file


    Parameters
    ----------
    BaseModel : _type_
        _description_
    """

    deployments: List[Deployment] = []
    latest_deployment: datetime.datetime = Field(default_factory=datetime.datetime.now)


class Build(BaseModel):
    manifest: Manifest
    build_id: str
    builder: str
    build_at: datetime.datetime = Field(default_factory=datetime.datetime.now)


class BuildsConfigFile(BaseModel):
    builds: List[Build] = Field(default_factory=list)
    latest_build: Optional[Build]



from fakts.fakts import Fakts
from fakts.fakts import Fakts
from typing import Optional
from fakts.grants.remote.discovery.well_known import WellKnownDiscovery
from fakts.grants.remote import RemoteGrant
from fakts.grants.remote.demanders.auto_save import AutoSaveDemander
from fakts.grants.remote.demanders.cache import AutoSaveCacheStore
from fakts.grants.remote.demanders.static import StaticDemander
from fakts.grants.remote.demanders.device_code import DeviceCodeDemander
from arkitekt.model import Manifest


class ArkitektFaktsQt(Fakts):
    grant: RemoteGrant


class ArkitektFakts(Fakts):
    pass


def build_arkitekt_fakts(
    manifest: Manifest,
    url: Optional[str] = None,
    no_cache: bool = False,
    headless: bool = False,
) -> ArkitektFakts:
    identifier = manifest.identifier
    version = manifest.version

    if no_cache:
        demander = DeviceCodeDemander(
            manifest=manifest,
            redirect_uri="http://localhost:6767",
            open_browser=not headless,
        )

    else:
        demander = AutoSaveDemander(
            demander=DeviceCodeDemander(
                manifest=manifest,
                redirect_uri="http://localhost:6767",
                open_browser=not headless,
            ),
            store=AutoSaveCacheStore(
                cache_file=f".arkitekt/cache/{identifier}-{version}_fakts_cache.json"
            ),
        )

    return ArkitektFakts(
        grant=RemoteGrant(
            demander=demander,
            discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
        )
    )


def build_arkitekt_token_fakts(
    manifest: Manifest,
    token: str,
    url,
    no_cache: Optional[bool] = False,
    headless=False,
):
    return ArkitektFakts(
        grant=RemoteGrant(
            demander=StaticDemander(token=token),
            discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
        )
    )



from arkitekt.model import Manifest
from arkitekt.apps.types import QtApp
from koil.composition.qt import QtPedanticKoil
from arkitekt.apps.fallbacks import ImportException, InstallModuleException
from typing import Any, Optional


def build_arkitekt_qt_app(
    manifest: Manifest,
    no_cache: bool = False,
    instance_id: Optional[str] = None,
    beacon_widget: Any = None,
    login_widget: Any = None,
    parent: Any = None,
    settings: Any = None,
):
    if settings is None:
        try:
            from qtpy import QtCore

            settings = QtCore.QSettings()
        except ImportError as e:
            raise InstallModuleException(
                "Please install qtpy to use arkitekt_qt"
            ) from e

    try:
        from arkitekt.apps.service.fakts_qt import build_arkitekt_qt_fakts

        fakts = build_arkitekt_qt_fakts(
            manifest=manifest,
            no_cache=no_cache,
            beacon_widget=beacon_widget,
            parent=parent,
            settings=settings,
        )
    except ImportError as e:
        fakts = ImportException(import_exception=e, install_library="qtpy")

    try:
        from arkitekt.apps.service.herre_qt import build_arkitekt_qt_herre

        herre = build_arkitekt_qt_herre(
            manifest=manifest,
            fakts=fakts,
            login_widget=login_widget,
            parent=parent,
            settings=settings,
        )
    except ImportError as e:
        herre = ImportException(import_exception=e, install_library="qtpy")

    try:
        from arkitekt.apps.service.rekuest import build_arkitekt_rekuest

        rekuest = build_arkitekt_rekuest(
            fakts=fakts, herre=herre, instance_id=instance_id or "main"
        )
    except ImportError as e:
        rekuest = ImportException(import_exception=e, install_library="rekuest")

    try:
        from arkitekt.apps.service.mikro import build_arkitekt_mikro

        mikro = build_arkitekt_mikro(fakts=fakts, herre=herre)
    except ImportError as e:
        mikro = ImportException(import_exception=e, install_library="mikro")

    try:
        from arkitekt.apps.service.unlok import build_arkitekt_unlok

        unlok = build_arkitekt_unlok(herre=herre, fakts=fakts)
    except ImportError as e:
        unlok = ImportException(import_exception=e, install_library="unlok")

    try:
        from arkitekt.apps.service.fluss import build_arkitekt_fluss

        fluss = build_arkitekt_fluss(herre=herre, fakts=fakts)
    except ImportError as e:
        fluss = ImportException(import_exception=e, install_library="fluss")

    return QtApp(
        koil=QtPedanticKoil(parent=parent),
        manifest=manifest,
        fakts=fakts,
        herre=herre,
        rekuest=rekuest,
        mikro=mikro,
        unlok=unlok,
        fluss=fluss,
    )



import rich_click as click
from arkitekt.cli.vars import get_manifest
from arkitekt.cli.types import Requirement
import subprocess
from click import Context
from arkitekt.cli.io import get_builds


@click.command()
@click.option("--build", help="The build to use", type=str, default="latest")
@click.option(
    "--url", help="The fakts server to use", type=str, default="localhost:8000"
)
@click.option(
    "--builder", help="The builder to use", type=str, default="arkitekt.builders.easy"
)
@click.pass_context
def stage(ctx: Context, build: str, url: str, builder: str) -> None:
    """Stages the latest Build for testing

    Stages the current build for testing. This will create a temporary staged version
    of the app that is run agains the local arkitekt instance. The builder will be changed
    to the easy or provided builder to ensure that the app can be run headlessly


    """

    manifest = get_manifest(ctx)

    builds = get_builds()
    assert (
        build in builds
    ), f"Build {build} not found. Please run `arkitekt build` first"
    build_model = builds[build]
    build_id = build_model.build_id

    base_command = ["docker", "run", "-it"]
    with_network = ["--net", "host"]
    with_gpus = ["--gpus", "all"]
    with_build_id = [build_id]
    with_builder = ["arkitekt", "run", "prod", "--builder", builder, "--headless"]
    with_url = ["--url", url]

    # Bulding the command
    command = base_command
    command += with_network

    if Requirement.GPU.value in manifest.requirements:
        click.echo("GPU Requirement found. Staging with GPU")
        command += with_gpus

    command += with_build_id
    command += with_builder
    command += with_url

    click.echo(f"Running inside docker: {manifest.identifier}:{manifest.version}")
    docker_run = subprocess.run(command)

    message_string = (
        docker_run.stdout.decode("utf-8")
        if docker_run.stdout
        else "" + docker_run.stderr.decode("utf-8")
        if docker_run.stderr
        else ""
    )

    if "No manifest found" in message_string:
        raise click.ClickException(
            "Looks like the docker container could not find the manifest. Did you mount the '.arkitekt folder' correctly?"
        )

    raise click.ClickException("Docker container exited")



import rich_click as click
from rich.table import Table
from rich.panel import Panel
from rich.console import Group
from arkitekt.cli.vars import *
from arkitekt.cli.constants import *
from arkitekt.cli.io import write_manifest


@click.group("scopes")
@click.pass_context
def scopes_group(ctx):
    """Inspect, add and remove scopes to this arkitekt app

    Scopes are rights that are granted to any arkitekt application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """
    pass


@scopes_group.command("add")
@click.argument(
    "SCOPE",
    nargs=-1,
    type=click.Choice(compile_scopes()),
)
@click.pass_context
def add_scopes(ctx, scope):
    """ "Acd scopes

    Scopes are rights that are granted to any arkitekt application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """
    if not scope:
        raise click.ClickException("Please provide at least one scope")

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    if scope:
        manifest.scopes = set(list(scope) + manifest.scopes)
        write_manifest(manifest)
        console.print(f"Scopes Updated to {manifest.scopes}")


@scopes_group.command("remove")
@click.argument(
    "SCOPE",
    nargs=-1,
    type=click.Choice(compile_scopes()),
)
@click.pass_context
def remove_scopes(ctx, scope):
    """Remove scopes

    Scopes are rights that are granted to any arkitekt application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """
    if not scope:
        raise click.ClickException("Please provide at least one scope to remove")

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    if scope:
        manifest.scopes = set(manifest.scopes) - set(scope)
        write_manifest(manifest)
        console.print(f"Scopes Updated to {manifest.scopes}")


@scopes_group.command("list")
@click.pass_context
def list_scopes(ctx):
    """List all the [i] currently [/] active scopes

    Scopes are rights that are granted to any arkitekt application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    table = Table.grid()
    table.padding = (0, 1)
    table.add_column("Scope")
    table.add_column("Description")
    for scope in manifest.scopes:
        table.add_row(scope, "TODO: This should be a description")

    panel = Panel(
        Group("[bold green]Demanded Scopes[/]", table),
        title_align="center",
        border_style="green",
        style="white",
    )

    console.print(panel)


@scopes_group.command("available")
@click.pass_context
def list_available(ctx):
    """List all the [i] available [/]  scopes

    Scopes are rights that are granted to any arkitekt application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """

    console = get_console(ctx)

    table = Table.grid()
    table.padding = (0, 1)
    table.add_column("Scope")
    table.add_column("Description")
    for scope in compile_scopes():
        table.add_row(scope, "TODO")

    panel = Panel(
        Group("[bold green]Available Scopes[/]", table),
        title_align="center",
        border_style="green",
        style="white",
    )

    console.print(panel)



import rich_click as click
from arkitekt.cli.vars import get_manifest, get_console
from arkitekt.cli.constants import *
from rich.table import Table
from rich.panel import Panel
from rich.console import Group
from arkitekt.cli.vars import *
from arkitekt.cli.constants import *
from arkitekt.cli.io import write_manifest


@click.group("requirements")
@click.pass_context
def requirements_group(ctx):
    """Inspect, add and remove requirements to this arkitekt app

    Requirements are used to hint the user to have specific hardware or software installed, when
    using your application. It is also used to inform the platform about the requirements of your app,
    when installing it in its plugin sandbox. for more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]

    """


@requirements_group.command("add")
@click.argument(
    "REQUIREMENTS",
    nargs=-1,
    type=click.Choice(compile_requirements()),
)
@click.pass_context
def add_requirement(ctx, requirements):
    """Add requiremenets

    Requirements are used to hint the user to have specific hardware or software installed, when
    using your application. It is also used to inform the platform about the requirements of your app,
    when installing it in its plugin sandbox. for more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]

    """
    if not requirements:
        raise click.ClickException("Please provide at least one requirement")

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    if requirements:
        manifest.requirements = set(list(requirements) + manifest.requirements)
        write_manifest(manifest)
        console.print(f"Requirements Updated to {manifest.requirements}")


@requirements_group.command("remove")
@click.argument(
    "REQUIREMENTS",
    nargs=-1,
    type=click.Choice(compile_requirements()),
)
@click.pass_context
def remove_requirements(ctx, requirements):
    """Remove requirements

    Requirements are used to hint the user to have specific hardware or software installed, when
    using your application. It is also used to inform the platform about the requirements of your app,
    when installing it in its plugin sandbox. for more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """
    if not requirements:
        raise click.ClickException("Please provide at least one requirement to remove")

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    if requirements:
        manifest.requirements = set(manifest.requirements) - set(requirements)
        write_manifest(manifest)
        console.print(f"Requirements Updated to {manifest.requirements}")


@requirements_group.command("list")
@click.pass_context
def list_requirements(ctx):
    """Lists the [i]current[/] requirements

    Requirements are used to hint the user to have specific hardware or software installed, when
    using your application. It is also used to inform the platform about the requirements of your app,
    when installing it in its plugin sandbox. for more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]

    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    table = Table(
        title="[green bold ]Requirements[/]",
        title_justify="left",
        title_style="green bold",
    )
    table.add_column("Requirement")
    table.add_column("Description")

    for scope in manifest.requirements:
        table.add_row(scope, "TODO")

    panel = Panel(
        table,
        title_align="center",
        border_style="green",
        style="white",
    )

    console.print(panel)


@requirements_group.command("available")
@click.pass_context
def list_available_requirements(ctx):
    """Lists the [i]available[/] requirements

    Requirements are used to hint the user to have specific hardware or software installed, when
    using your application. It is also used to inform the platform about the requirements of your app,
    when installing it in its plugin sandbox. for more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]

    """

    console = get_console(ctx)

    table = Table.grid()
    table.add_column("Scope")
    table.add_column("Description")
    for scope in compile_requirements():
        table.add_row(scope, "TODO")

    panel = Panel(
        Group("[bold green]Available Scopes[/]", table),
        title_align="center",
        border_style="green",
        style="white",
    )

    console.print(panel)



from typing import List, Optional
import logging
import os
from .utils import create_arkitekt_folder
from .model import Manifest
from arkitekt.apps.types import NextApp, EasyApp
from arkitekt.apps.easy import build_arkitekt_easy_app
from arkitekt.apps.qt import build_arkitekt_qt_app
from arkitekt.apps.fallbacks import InstallModuleException


def easy(
    identifier: str,
    version: str = "0.0.1",
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    url: str = "http://localhost:8000",
    headless: bool = False,
    log_level: str = "ERROR",
    token: Optional[str] = None,
    no_cache: bool = False,
    instance_id: str = "main",
    register_reaktion: bool = False,
    app_kind: str = "development",
    enforce: Optional[List[str]] = None,
) -> EasyApp:
    """Creates an easy app

    A simple way to create an Arkitekt app, that will be able to connect to the
    Arkitekt server in a variety of scenarious. It should be used for most
    Arkitekt apps, as it will provide access to the most common features and
    services.

    A few things to note:
        - Easy apps will try to establish themselves a "development" apps, by default
          which means that they will be authenticated with the Arkitekt server on
          a per user basis. If you want to create a "desktop" app, which multiple users
          can use, you should set the `app_kind` to "desktop" TODO: Currently not implemented (use next app for this)
        - The Easy builder can also be used in plugin apps, and when provided with a fakts token
           will be able to connect to the Arkitekt server without any user interaction.


    Parameters
    ----------
    identifier : str
        The apps identifier (should be globally unique, see Manifest for more info)
    version : str, optional
        The version of the app, by default "0.0.1"
    logo : str, optional
        The logo of the app as a public http url, by default None
    scopes : List[str], optional
        The scopes, that this apps requires, will default to standard scopes, by default None
    url : str, optional
        The fakts server that will be used to configure this app, in a default Arkitekt deployment this
        is the address of the "Lok Service" (which provides the Fakts API), by default "http://localhost:8000"
        Will be overwritten by the FAKTS_URL environment variable
    headless : bool, optional
        Should we run in headless, mode, e.g printing necessary interaction into the console (will forexample
        stop opening browser windows), by default False
    log_level : str, optional
        The log-level to use, by default "ERROR"
    token : str, optional
        A fakts token to use, by default None
        Will be overwritten by the FAKTS_TOKEN environment variable
    no_cache : bool, optional
        Should we skip caching token, acess-token, by default False
        Attention: If this is set to True, the app will always have to be configured
        and authenticated.
    instance_id : str, optional
        The instance_id to use, by default "main"
        Can be set to a different value, if you want to run multiple intstances
        of the same app by the same user.
        Will be overwritten by the REKUEST_INSTANCE_ID environment variable
    register_reaktion : bool, optional
        Should we register the reaktion extension, by default True
        If set to False, the app will not be able to use the reaktion extension
        (which is necessary for scheduling in app` workflows from fluss)
    app_kind : str, optional
        The kind of app to create, by default "development"
        Can be set to "desktop" to create a desktop app, that can be used by multiple users.
        While this is currently not implemented, the next app will be able to do this.

    Returns
    -------
    EasyApp
        A built app, that can be used to interact with the Arkitekt server
    """

    url = os.getenv("FAKTS_URL", url)
    token = os.getenv("FAKTS_TOKEN", token)
    instance_id = os.getenv("REKUEST_INSTANCE_ID", instance_id)

    create_arkitekt_folder(with_cache=True)

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
    )

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = build_arkitekt_easy_app(
        manifest=manifest,
        url=url,
        no_cache=no_cache,
        headless=headless,
        instance_id=instance_id,
        token=token,
        enforce=enforce,
    )

    if register_reaktion:
        try:
            from reaktion.extension import ReaktionExtension

            app.rekuest.agent.extensions["reaktion"] = ReaktionExtension()
        except ImportError as e:
            raise InstallModuleException(
                "You need to install reaktion to use the reaktion extension"
            ) from e

    return app


def next(
    identifier: str,
    version: str = "latest",
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    url: str = "http://localhost:8000",
    headless: bool = False,
    log_level: str = "ERROR",
    token: Optional[str] = None,
    no_cache: bool = False,
    instance_id: str = "main",
    register_reaktion: bool = False,
    app_kind: str = "development",
) -> NextApp:
    """Creates a next app

    A simple way to create an Arkitekt Next app, Arkitekt next apps are
    development apps by default, as they will try to register themselves
    with services that are not yet available in production (such as the
    rekuest_next and mikro_next services). They represent the next generation
    of Arkitekt apps, and will be the default way to create Arkitekt apps
    in the future. From here be dragons.

    A few things to note:
        -   The Next builder closely mimics the easy builder, but will use the
            next generation of services (such as rekuest_next and mikro_next)
            and will therefore not be compatible with the current generation.

        -  Next apps will try to establish themselves a "development" apps, by default
            which means that they will be authenticated with the Arkitekt server on
            a per user basis. If you want to create a "desktop" app, which multiple users
            can use, you should set the `app_kind` to "desktop" TODO: Currently not implemented (use next app for this)
        -  The Next builder can also be used in plugin apps, and when provided with a fakts token
           will be able to connect to the Arkitekt server without any user interaction.


    Parameters
    ----------
    identifier : str
        The apps identifier (should be globally unique, see Manifest for more info)
    version : str, optional
        The version of the app, by default "0.0.1"
    logo : str, optional
        The logo of the app as a public http url, by default None
    scopes : List[str], optional
        The scopes, that this apps requires, will default to standard scopes, by default None
    url : str, optional
        The fakts server that will be used to configure this app, in a default Arkitekt deployment this
        is the address of the "Lok Service" (which provides the Fakts API), by default "http://localhost:8000"
        Will be overwritten by the FAKTS_URL environment variable
    headless : bool, optional
        Should we run in headless, mode, e.g printing necessary interaction into the console (will forexample
        stop opening browser windows), by default False
    log_level : str, optional
        The log-level to use, by default "ERROR"
    token : str, optional
        A fakts token to use, by default None
        Will be overwritten by the FAKTS_TOKEN environment variable
    no_cache : bool, optional
        Should we skip caching token, acess-token, by default False
        Attention: If this is set to True, the app will always have to be configured
        and authenticated.
    instance_id : str, optional
        The instance_id to use, by default "main"
        Can be set to a different value, if you want to run multiple intstances
        of the same app by the same user.
        Will be overwritten by the REKUEST_INSTANCE_ID environment variable
    register_reaktion : bool, optional
        Should we register the reaktion extension, by default True
        If set to False, the app will not be able to use the reaktion extension
        (which is necessary for scheduling in app` workflows from fluss)
    app_kind : str, optional
        The kind of app to create, by default "development"
        Can be set to "desktop" to create a desktop app, that can be used by multiple users.

    Returns
    -------
    NextApp
        A built app, that can be used to interact with the Arkitekt server
    """
    from arkitekt.apps.next import build_next_app

    url = os.getenv("FAKTS_URL", url)
    token = os.getenv("FAKTS_TOKEN", token)

    create_arkitekt_folder(with_cache=True)

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
    )

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = build_next_app(
        manifest=manifest,
        url=url,
        no_cache=no_cache,
        headless=headless,
        instance_id=instance_id,
        token=token,
        app_kind=app_kind,
    )

    if register_reaktion:
        try:
            from reaktion.extension import ReaktionExtension

            app.rekuest.agent.extensions["reaktion"] = ReaktionExtension()
        except ImportError as e:
            raise InstallModuleException(
                "You need to install reaktion to use the reaktion extension"
            ) from e

    return app


def jupy(
    identifier: str,
    version: str = "0.0.1",
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    url: str = "http://localhost:8000",
    headless: bool = False,
    log_level: str = "ERROR",
    token: Optional[str] = None,
    no_cache: bool = False,
    instance_id: str = "main",
    register_reaktion: bool = False,
    app_kind: str = "development",
):
    """Creates a Jupyter app

    A simple way to create an Arkitekt app, that will be able to connect to the
    Arkitekt server when running inside a jupyter notebook. It simply wraps the
    the easy app, and sets the allow_sync_in_async flag to True, so that the
    the app can reliably be used in jupyter notebooks (which are async by default,
    but often are used in a sync manner). Additionally, it will also auto_enter
    the app, so that you dont have to use the app in a with statement.

    Attention:
        Running Arkitekt apps in sync mode comes with a few caveats, as the
        app will be put into another thread, and when using the app without a
        a context manager (the with statement), you will have to manually
        exit the app, by calling the `app.exit()` method. Arkitekt will otherwise
        not be able to properly shut down the app, and will continue to run in
        the background (until the notebook is restarted)

        This should only be a big problem, if you want to run multiple apps
        in the same notebook, and therefore should not be necessary in most cases,
        but you should be aware of this.


    A few things to note:
        -  Jupy apps will try to establish themselves a "development" apps, by default
            which means that they will be authenticated with the Arkitekt server on
            a per user basis. If you want to create a "desktop" app, which multiple users
            can use, you should set the `app_kind` to "desktop" TODO: Currently not implemented (use next app for this)
        -  The Jupy builder can also be used in plugin apps, and when provided with a fakts token
           will be able to connect to the Arkitekt server without any user interaction.


    Parameters
    ----------
    identifier : str
        The apps identifier (should be globally unique, see Manifest for more info)
    version : str, optional
        The version of the app, by default "0.0.1"
    logo : str, optional
        The logo of the app as a public http url, by default None
    scopes : List[str], optional
        The scopes, that this apps requires, will default to standard scopes, by default None
    url : str, optional
        The fakts server that will be used to configure this app, in a default Arkitekt deployment this
        is the address of the "Lok Service" (which provides the Fakts API), by default "http://localhost:8000"
        Will be overwritten by the FAKTS_URL environment variable
    headless : bool, optional
        Should we run in headless, mode, e.g printing necessary interaction into the console (will forexample
        stop opening browser windows), by default False
    log_level : str, optional
        The log-level to use, by default "ERROR"
    token : str, optional
        A fakts token to use, by default None
        Will be overwritten by the FAKTS_TOKEN environment variable
    no_cache : bool, optional
        Should we skip caching token, acess-token, by default False
        Attention: If this is set to True, the app will always have to be configured
        and authenticated.
    instance_id : str, optional
        The instance_id to use, by default "main"
        Can be set to a different value, if you want to run multiple intstances
        of the same app by the same user.
        Will be overwritten by the REKUEST_INSTANCE_ID environment variable
    register_reaktion : bool, optional
        Should we register the reaktion extension, by default True
        If set to False, the app will not be able to use the reaktion extension
        (which is necessary for scheduling in app` workflows from fluss)
    app_kind : str, optional
        The kind of app to create, by default "development"
        Can be set to "desktop" to create a desktop app, that can be used by multiple users.

    Returns
    -------
    NextApp
        A built app, that can be used to interact with the Arkitekt server
    """
    app = easy(
        identifier=identifier,
        version=version,
        url=url,
        token=token,
        logo=logo,
        scopes=scopes,
        log_level=log_level,
        register_reaktion=register_reaktion,
        app_kind=app_kind,
        headless=headless,
        instance_id=instance_id,
        no_cache=no_cache,
    )
    app.koil.sync_in_async = True

    if register_reaktion:
        try:
            from reaktion.extension import ReaktionExtension

            app.rekuest.agent.extensions["reaktion"] = ReaktionExtension()
        except ImportError as e:
            raise InstallModuleException(
                "You need to install reaktion to use the reaktion extension"
            ) from e

    app.enter()  # This will start the event loop with the sync interface (for automatic awaits in jupyter)
    return app


def port(**kwargs):
    """Easy port creation

    A simple way to create an Arkitekt app with a device code grant
    it will cache the configuration in a file called `identifier`_token.json
    and the cache in a file called `identifier`_cache.json

    Args:
        identifier (str): The apps identifier
        version (str, optiosnal): The apps verion. Defaults to "latest".
        url (_type_, optional): The app configuration url. Defaults to "http://localhost:8000/f/".
        headless (bool, optional): Do not open a browser window. Defaults to False.
        allow_sync_in_async (bool, optional): Should we allow the creation of a sync interface in an async loop (necessary for a sync interface in jupyter). Defaults to True.

    Returns:
        Arkitekt: _description_
    """

    if not kwargs.get("token", None):
        raise ValueError("You must provide a token")

    if not kwargs.get("url", None):
        raise ValueError("You must provide a url")

    return easy(**kwargs)


def scheduler(
    identifier: str,
    version: str = "latest",
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    url: str = "http://localhost:8000",
    headless: bool = False,
    log_level: str = "ERROR",
    token: Optional[str] = None,
    no_cache: bool = False,
    instance_id: str = "main",
):
    """Scheduler app creation

    A an arkitekt scheduler app with a device code grant and a reaktion agent
    this app will be able to deploy graphs from fluss

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        url (_type_, optional): The app configuration url. Defaults to "http://localhost:8000/f/".
        headless (bool, optional): Do not open a browser window. Defaults to False.
        allow_sync_in_async (bool, optional): Should we allow the creation of a sync interface in an async loop (necessary for a sync interface in jupyter). Defaults to True.

    Returns:
        Arkitekt: _description_
    """

    try:
        from reaktion.extension import ReaktionExtension
    except ImportError as e:
        raise ImportError(
            "You need to install reaktion to use the scheduler function"
        ) from e

    url = os.getenv("FAKTS_URL", url)
    token = os.getenv("FAKTS_TOKEN", token)

    create_arkitekt_folder(with_cache=True)

    Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
    )

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = easy(
        identifier=identifier,
        version=version,
        url=url,
        token=token,
        logo=logo,
        scopes=scopes,
        log_level=log_level,
        headless=headless,
        instance_id=instance_id,
        no_cache=no_cache,
    )
    app.rekuest.agent.extensions["reaktion"] = ReaktionExtension()

    return app


def publicqt(
    identifier: str,
    version: str = "latest",
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    url: str = "http://localhost:8000",
    headless: bool = False,
    log_level: str = "ERROR",
    token: Optional[str] = None,
    no_cache: bool = False,
    instance_id: str = "main",
    parent=None,
    beacon_widget=None,
    login_widget=None,
    force_herre_grant=None,
    settings=None,
):
    """Public QtApp creation

    A simple way to create an Arkitekt app with a public grant (allowing users to sign
    in with the application ) utlizing a retrieve grant (necessating a previous configuration
    of the application on the server side)

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        parent (QtWidget, optional): The QtParent (for the login and server select widget). Defaults to None.

    Returns:
        Arkitekt: The Arkitekt app
    """

    from arkitekt.apps.qt import build_arkitekt_qt_app

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes or ["openid", "read", "write"],
        logo=logo,
    )

    create_arkitekt_folder(with_cache=True)

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = build_arkitekt_qt_app(
        manifest=manifest,
        parent=parent,
        no_cache=no_cache,
        beacon_widget=beacon_widget,
        login_widget=login_widget,
        instance_id=instance_id,
        settings=settings,
    )
    print("Entering", parent)
    app.enter()  # This will start the event loop and attach the event handlers
    return app


def publicscheduleqt(
    identifier: str,
    version: str = "latest",
    parent=None,
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    beacon_widget=None,
    login_widget=None,
    force_herre_grant=None,
    instance_id: str = "main",
    no_cache: bool = False,
    log_level: str = "ERROR",
    settings=None,
):
    """Public QtApp creation

    A simple way to create an Arkitekt app with a public grant (allowing users to sign
    in with the application ) utlizing a retrieve grant (necessating a previous configuration
    of the application on the server side)

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        parent (QtWidget, optional): The QtParent (for the login and server select widget). Defaults to None.

    Returns:
        Arkitekt: The Arkitekt app
    """
    try:
        from reaktion.extension import ReaktionExtension
    except ImportError as e:
        raise ImportError(
            "You need to install reaktion to use the scheduler function"
        ) from e

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes or ["openid", "read", "write"],
        logo=logo,
    )

    create_arkitekt_folder(with_cache=True)

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = build_arkitekt_qt_app(
        manifest=manifest,
        parent=parent,
        no_cache=no_cache,
        beacon_widget=beacon_widget,
        login_widget=login_widget,
        instance_id=instance_id,
        settings=settings,
    )
    app.rekuest.agent.extensions["reaktion"] = ReaktionExtension()
    app.enter()
    return app



import rich_click as click
from arkitekt.cli.vars import get_console, get_manifest
import os
from rich.panel import Panel
import subprocess
import uuid
from arkitekt.cli.io import generate_build
from click import Context

@click.command()
@click.option("--dockerfile", help="The dockerfile to use", default="Dockerfile")
@click.option(
    "--builder", help="The port builder to use", default="arkitekt.builders.port"
)
@click.pass_context
def build(ctx: Context, dockerfile: str, builder: str) -> None:
    """Builds the arkitekt app to docker"""

    build_id = str(uuid.uuid4())

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    if not os.path.exists(dockerfile):
        raise click.ClickException(
            f"Dockerfile {dockerfile} does not exist. Please create a dockerfile first"
             "(e.g. with the port wizard command)."
        )

    md = Panel(
        "Building for Port", subtitle="This may take a while...", subtitle_align="right"
    )

    md = Panel("Building Docker Container")
    console.print(md)

    docker_run = subprocess.run(
        ["docker", "build", "-t", build_id, "-f", dockerfile or "Dockerfile", "."]
    )
    if docker_run.returncode != 0:
        raise click.ClickException("Could not build docker container")

    generate_build(builder, build_id, manifest)



from typing import Any
from koil.composition import KoiledModel


class InstallModuleException(Exception):
    """An exception that is raised when a module is not installed."""

    pass


class ImportException(KoiledModel):
    """A pydantic model that raises an InstallModuleException when an attribute
    of a non-installed library is accessed.

    We use this to allow for optional imports of libraries that are not
    required for the core functionality of arkitekt, but that are required
    for some of the features. This allows us to install arkitekt without
    having to install all the dependencies. If a feature is used that requires
    a library that is not installed, we raise an InstallModuleException.

    Check the app builders for examples of how this is used.

    TODO: Maybe we should write better documentation for this.
    """

    import_exception: Exception
    install_library: str

    def __getattr__(self, __name: str) -> Any:
        """Override the __getattribute__ method to raise an InstallModuleException"""
        try:
            raise self.import_exception
        except Exception as e:
            raise InstallModuleException(
                f"This module is not installed! Please install the library {self.install_library} to use this feature. Cannot access attribute {__name}"
            ) from e

    class Config:
        arbitrary_types_allowed = True



import os
import shutil
import rich_click as click
from arkitekt.cli.options import *


@click.command()
@with_boring
@with_choose_services
@with_graphql_config
@with_api_path
@with_graphql_config
@with_docuements
@with_schemas
def init(boring, service, config, documents, schemas, overwrite_config, path):
    """Initialize code generation for the arkitekt app

    Code generation for API's is done with the help of GraphQL Code Generation
    that is powered by turms. This command initializes the code generation for
    the app. It creates the necessary folders and files for the code generation
    to work. It also creates a graphql config file that is used by turms to
    generate the code.

    """
    app_directory = os.getcwd()

    if documents:
        os.makedirs(os.path.join(app_directory, "documents"), exist_ok=True)
    if schemas:
        os.makedirs(os.path.join(app_directory, "schemas"), exist_ok=True)
    if path:
        os.makedirs(os.path.join(app_directory, path), exist_ok=True)

    # Initializing the config
    projects = {}

    for service, version in service:
        config_path = build_relative_dir("configs", service, f"{version}.yaml")
        documents_path = build_relative_dir("documents", service, version)
        schema_path = build_relative_dir("schemas", service, f"{version}.graphql")

        if schemas:
            if os.path.exists(schema_path):
                try:
                    shutil.copyfile(
                        schema_path,
                        os.path.join(app_directory, "schemas", f"{service}.graphql"),
                    )
                except FileExistsError:
                    if click.confirm(
                        f"Schema for {service} {version} already exists. Do you want to overwrite?"
                    ):
                        shutil.copytree(
                            documents_path,
                            os.path.join(app_directory, "documents", service),
                            dirs_exist_ok=True,
                        )
            else:
                get_console().print(f"[red]No schema found for {service} {version}[/]")

        if documents:
            if os.path.exists(documents_path):
                try:
                    shutil.copytree(
                        documents_path,
                        os.path.join(app_directory, "documents", service),
                    )
                except FileExistsError:
                    if click.confirm(
                        f"Documents for {service} {version} already exist. Do you want to overwrite them?"
                    ):
                        shutil.copytree(
                            documents_path,
                            os.path.join(app_directory, "documents", service),
                            dirs_exist_ok=True,
                        )

            else:
                get_console().print(f"[red]No schema found for {service} {version}[/]")

        project = yaml.load(open(config_path, "r"), Loader=yaml.FullLoader)
        if schemas:
            project["schema"] = "schemas/" + service + ".graphql"
        if documents:
            project["documents"] = "documents/" + service + "/**/*.graphql"

        project["extensions"]["turms"]["out_dir"] = path
        project["extensions"]["turms"]["generated_name"] = f"{service}.py"

        projects[service] = project

    if overwrite_config:
        graph_config_path = os.path.join(app_directory, config)
        yaml.safe_dump(
            {"projects": projects}, open(graph_config_path, "w"), sort_keys=False
        )
        get_console().print(f"Config file written to {graph_config_path}")



WELCOME_MESSAGE = (
    "Welcome to Arkitekt. Arkitekt is a bioimage analysis framework for building"
    " beautiful and fast interfaces around your python code. It is buid on top of "
    "composable packages like mikro, fluss and rekuest, which allow you to build simple"
    " apps with ease."
)


LOGO = """
            _    _ _       _    _   
  __ _ _ __| | _(_) |_ ___| | _| |_ 
 / _` | '__| |/ / | __/ _ \ |/ / __|
| (_| | |  |   <| | ||  __/   <| |_ 
 \__,_|_|  |_|\_\_|\__\___|_|\_\\__|
                                                                   
"""

ERROR_EPILOGUE = (
    "To find out more, visit [link=https://arkitekt.live]https://arkitekt.live[/link]"
)



import rich_click as click
import subprocess
from .utils import search_username_in_docker_info
from arkitekt.cli.vars import get_console
from arkitekt.cli.types import Manifest
from rich.panel import Panel
from arkitekt.cli.io import get_builds, get_deployments, generate_deployment
from click import Context

def check_if_manifest_already_deployed(manifest: Manifest) -> None:
    """Checks if a manifest has already been deployed. If it has, it raises a click.ClickException.

    Parameters
    ----------
    manifest : Manifest
        THe manifest to check

    Raises
    ------
    click.ClickException
        A click exception if the manifest has already been deployed
    """
    config = get_deployments()
    for deployment in config.deployments:
        if (
            deployment.manifest.identifier == manifest.identifier
            and deployment.manifest.version == manifest.version
        ):
            raise click.ClickException(
                f"Deployment of {manifest.identifier}/{manifest.version} already exists. You cannot deploy the same version twice."
            )

@click.command()
@click.option("--build", help="The build to use", type=str, default="latest")
@click.option("--tag", help="The tag to use")
@click.pass_context
def publish(ctx: Context, build: str , tag: str) -> None:
    """Deploys aa previous build to dockerhub"""

    console = get_console(ctx)

    builds = get_builds()
    assert (
        build in builds
    ), f"Build {build} not found. Please run `arkitekt build` first"
    build_model = builds[build]

    manifest = build_model.manifest
    if build_model.manifest.version == "dev":
        raise click.ClickException(
            "You cannot deploy a dev version. Please run `arkitekt version` first to set a version"
        )

    check_if_manifest_already_deployed(manifest)
    docker_info = subprocess.check_output(["docker", "info"]).decode("utf-8")
    username = search_username_in_docker_info(docker_info)
    if not username:
        username = click.prompt(
            "Could not find username in docker info. Please provide your docker username"
        )

    tag = tag or click.prompt(
        "The tag to use",
        default=f"{username}/{build_model.manifest.identifier}:{build_model.manifest.version}",
    )

    md = Panel("Building Docker Container")
    console.print(md)

    deployed = {}

    docker_run = subprocess.run(["docker", "tag", build_model.build_id, tag])
    if docker_run.returncode != 0:
        raise click.ClickException("Could not retag docker container")

    console.print(md)
    docker_run = subprocess.run(["docker", "push", tag])
    if docker_run.returncode != 0:
        raise click.ClickException("Could not push docker container")

    deployed["docker"] = tag

    generate_deployment(
        build_model,
        tag,
        with_definitions=False,
    )

    md = Panel(
        "[bold green] Sucessfully pushed your plugin to dockerhub. Make sure to commit and push your changes to github to make them available to others.",
        title="Yeah! Success",
        title_align="center",
        border_style="green",
        style="green",
    )
    console.print(md)



from pydantic import BaseModel, Field
from typing import List, Optional


class Manifest(BaseModel):
    """A manifest for an app that can be installed in Arkitekt

    Manifests are used to describe apps that can be installed in Arkitekt.
    They provide information about the app, such as the
    its globally unique identifier, the version, the scopes it needs, etc.

    This Manifest is send to the Fakts server on initial app configuration,
    and is used to register the app with the Fakts server, which in turn
    will prompt the user to grant the app access to establish itself as
    an Arkitekt app (and therefore as an OAuth2 client) (see more in the
    Fakts documentation).

    """

    version: str
    """ The version of the app TODO: Should this be a semver? """
    identifier: str
    """ The globally unique identifier of the app: TODO: Should we check for a reverse domain name? """
    scopes: List[str]
    """ Scopes that this app should request from the user """
    logo: Optional[str]
    """ A URL to the logo of the app TODO: We should enforce this to be a http URL as local paths won't work """

    class Config:
        extra = "forbid"


class User(BaseModel):
    """A user of Arkitekt

    This model represents a user on Arkitekt. As herre is acgnostic to the
    user model, we need to provide a model that can be used to represent
    the Arkitekt user. This model is used by the
    :class:`herre.fakts.fakts_endpoint_fetcher.FaktsUserFetcher` to
    fetch the user from the associated Arkitekt Lok instance. This model
    is closely mimicking the OIDC user model, and is therefore compatible
    to represent OIDC users.

    """

    id: str = Field(alias="sub")
    """ The user's id (in lok, this is the user's sub(ject) ID)"""

    username: str = Field(alias="preferred_username")
    """ The user's preferred username """
    email: str = Field(alias="email")
    """ The user's preferred username """



from arkitekt.utils import create_arkitekt_folder
import os
from typing import Optional
from arkitekt.cli.types import (
    Manifest,
    Build,
    BuildsConfigFile,
    Deployment,
    DeploymentsConfigFile,
)
import yaml
import json
from typing import Dict
import datetime


def load_manifest_yaml(path: str) -> Manifest:
    """Loads a manifest from a yaml file

    Uses yaml safe load to load the manifest from a yaml file
    (to avoid unsafe yaml attributes)

    Parameters
    ----------
    path : str
        The path to the yaml file

    Returns
    -------
    Manifest
        The loaded manifest
    """
    with open(path, "r") as file:
        manifest = yaml.safe_load(file)
        return Manifest(**manifest)


def load_manifest() -> Optional[Manifest]:
    """Loads the manifest from the arkitekt folder

    Will load the manifest from the current working directories
    arkitekt folder. If no folder exists, it will create one, but
    will not create a manifest.

    Returns
    -------
    Optional[Manifest]
        The loaded manifest, or None if no manifest exists
    """
    path = create_arkitekt_folder()
    config_file = os.path.join(path, "manifest.yaml")
    if os.path.exists(config_file):
        return load_manifest_yaml(config_file)
    return None


def write_manifest(manifest: Manifest):
    """Writes a manifest to the arkitekt folder

    Will write a manifest to the current working directories
    arkitekt folder. If no folder exists, it will create one.


    Parameters
    ----------
    manifest : Manifest
        The manifest to write
    """
    path = create_arkitekt_folder()
    config_file = os.path.join(path, "manifest.yaml")

    with open(config_file, "w") as file:
        yaml.safe_dump(
            json.loads(manifest.json(exclude_none=True, exclude_unset=True)),
            file,
            sort_keys=True,
        )


def get_builds() -> Dict[str, Build]:
    """Will load the builds.yaml file and return a dictionary of builds

    Will load the builds.yaml file and return a dictionary of builds
    where the key is the build_id and the value is the build object.


    Returns
    -------
    Dict[str, Build]
        The loaded builds
    """
    path = create_arkitekt_folder()
    config_file = os.path.join(path, "builds.yaml")

    builds = {}

    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            config = BuildsConfigFile(**yaml.safe_load(file))

            builds = {build.build_id: build for build in config.builds}
            if config.latest_build:
                builds["latest"] = config.latest_build
            return builds
    else:
        return {}


def generate_build(
    builder: str,
    build_id: str,
    manifest: Manifest,
) -> Build:
    """Generates a build from a builder, build_id and manifest

    Will generate a build from a builder, build_id and manifest,
    and write it to the builds.yaml file in the arkitekt folder.


    Parameters
    ----------
    builder : str
        The builder that was used to build the build
    build_id : str
        The build_id of the build
    manifest : Manifest
        The manifest of the build

    Returns
    -------
    Build
        The generated build
    """
    path = create_arkitekt_folder()

    config_file = os.path.join(path, "builds.yaml")

    build = Build(
        manifest=manifest,
        builder=builder,
        build_id=build_id,
    )

    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            config = BuildsConfigFile(**yaml.safe_load(file))
            config.builds.append(build)
            config.latest_build = build
    else:
        config = BuildsConfigFile(builds=[build], latest_build=build)

    with open(config_file, "w") as file:
        yaml.safe_dump(
            json.loads(config.json(exclude_none=True, exclude_unset=True)),
            file,
            sort_keys=True,
        )

    return build


def get_deployments() -> DeploymentsConfigFile:
    """Loads the deployments.yaml file and returns the deployments

    Will load the deployments.yaml file and return the deployments
    as a DeploymentsConfigFile object. If no deployments.yaml file
    exists, it will return an empty DeploymentsConfigFile object.

    Returns
    -------
    DeploymentsConfigFile
        The deployments as a DeploymentsConfigFile object
    """
    path = create_arkitekt_folder()
    config_file = os.path.join(path, "deployments.yaml")
    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            return DeploymentsConfigFile(**yaml.safe_load(file))
    else:
        return DeploymentsConfigFile()


def generate_deployment(
    build: Build,
    image: str,
    with_definitions: bool =True,
) -> Deployment:
    """Generates a deployment from a build and an image

    Parameters
    ----------

    build : Build
        The build that should be deployed
    image: str
        The image that is the actuall deployment of the build
    with_definitions: bool:
        Should we generated and inspect definitions to bundle with
        the deployment?

    Returns:
    ------
    Deployment: The created deployment

    """

    path = create_arkitekt_folder()

    config_file = os.path.join(path, "deployments.yaml")

    deployment = Deployment(
        build_id=build.build_id,
        manifest=build.manifest,
        builder=build.builder,
        definitions=[],
        image=image,
    )

    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            config = DeploymentsConfigFile(**yaml.safe_load(file))
            config.deployments.append(deployment)
            config.latest_deployment = datetime.datetime.now()
    else:
        config = DeploymentsConfigFile(deployments=[deployment])
        config.latest_deployment = datetime.datetime.now()

    with open(config_file, "w") as file:
        yaml.safe_dump(
            json.loads(config.json(exclude_none=True)),
            file,
            sort_keys=True,
        )

    return deployment



import rich_click as click
import os


@click.command()
@click.argument("projects", default=None, required=False, nargs=-1)
@click.option(
    "--config",
    help="The config to use",
    type=click.Path(exists=True),
    default=None,
)
def compile(projects, config):
    """Genererate the code of a project"

    Uses a previously generated graphql-config.yaml file to generate the code for a or multiple projects.
    If no project is specified, all projects will be generated.


    """
    app_directory = os.getcwd()

    from turms.run import scan_folder_for_single_config, load_projects_from_configpath
    from turms.cli.main import generate_projects

    config = config or scan_folder_for_single_config(app_directory)
    if not config:
        raise click.ClickException(
            f"No config file found. Please run `arkitekt gen init` in {app_directory} to create a default config file or specify a config file with the --config flag"
        )

    parsing_projects = load_projects_from_configpath(config)
    if projects:
        parsing_projects = {
            key: value for key, value in parsing_projects.items() if key in projects
        }

    if not parsing_projects:
        raise click.ClickException(
            f"No projects found with the name '{projects}'. Available Projects: {', '.join(parsing_projects.keys())}"
        )

    generate_projects(parsing_projects, title="Arkitekt Compile")

    pass



from arkitekt.cli.types import Manifest, Packager
from arkitekt.cli.vars import get_manifest
import sys
import rich_click as click
from click import Context
import subprocess
from rich import get_console

try:
    pass
except ImportError as e:
    raise ImportError("Please install rekuest to use this feature") from e


import os



def get_base_prefix_compat() -> str:
    """Get base/real prefix, or sys.prefix if there is none."""
    return (
        getattr(sys, "base_prefix", None)
        or getattr(sys, "real_prefix", None)
        or sys.prefix
    )


def in_virtualenv() -> bool:
    """Returns True if in a virtualenv, otherwise False."""
    return get_base_prefix_compat() != sys.prefix


def pip_no_gpu(manifest: Manifest, python_version: str) -> str:
    """A function to generate a dockerfile for pip without gpu support

    Parameters
    ----------
    manifest : Manifest
        The manifest of the project
    python_version : str
        The python version to use

    Returns
    -------
    str
        The dockerfile
    """
    get_console().print("[blue]Setting up environment...[/blue]")
    if not in_virtualenv():
        click.confirm(
            "You are not in a virtual environment. Consider using a virtual environment. Do you want to continue?",
            abort=True,
        )

    if not os.path.exists(".requirements.txt") and click.confirm(
        "Do you want to freeze the dependencies the current interpreter?"
    ):
        subprocess.check_call("pip freeze > requirements.txt", shell=True)

    get_console().print("[blue]Generating Dockerfile...[/blue]")

    BASE_IMAGE = f"FROM python:{python_version}\n"
    PIP_APPENDIX = "# setup pip  environment\nCOPY ./requirements.txt /tmp/requirements.txt\nRUN pip install /tmp/requirements.yml\n"
    WORKDIR_APPENDIX = "RUN mkdir /app \nWORKDIR /app"

    dockerfile = BASE_IMAGE
    if click.confirm(
        "Do you want to install a requirements.txt with the dependencies with pip?"
    ):
        dockerfile += PIP_APPENDIX

    if click.confirm("Do you want to use the current working directory as the app?"):
        dockerfile += WORKDIR_APPENDIX

    return dockerfile


def poetry_no_gpu(manifest: Manifest, python_version: str) -> str:
    """A function to generate a dockerfile for pip without gpu support

    Parameters
    ----------
    manifest : Manifest
        The manifest of the project
    python_version : str
        The python version to use

    Returns
    -------
    str
        The dockerfile
    """
    get_console().print("[blue]Setting up environment...[/blue]")
    if not in_virtualenv():
        click.confirm(
            "You are not in a virtual environment. Consider using a virtual environment. Do you want to continue?",
            abort=True,
        )

    if click.confirm("Do you want to lock the current dependencies?"):
        subprocess.check_call("poetry lock", shell=True)

    get_console().print("[blue]Generating Dockerfile...[/blue]")

    BASE_IMAGE = f"FROM python:{python_version}\n"
    SETUP_POETRY = "# Install dependencies\nRUN pip install poetry rich\nENV PYTHONUNBUFFERED=1\n# Install Project files\nCOPY pyproject.toml /tmp\nCOPY poetry.lock /tmp\nRUN poetry config virtualenvs.create false\nWORKDIR /tmp\nRUN poetry install\n"
    WORKDIR_APPENDIX = "RUN mkdir /app\nCOPY . /app\nWORKDIR /app\n"

    dockerfile = BASE_IMAGE
    dockerfile += SETUP_POETRY

    if click.confirm("Do you want to use the current working directory as the app?"):
        dockerfile += WORKDIR_APPENDIX

    return dockerfile


CONDA_GPU = """
FROM jhnnsrs/tensorflow_conda_gpu:latest

# setup conda virtual environment
COPY ./requirements.yml /tmp/requirements.yml
RUN conda env create --name camera-seg -f /tmp/requirements.yml

RUN conda activate camera-seg   
"""


CONDA_NO_GPU = """
FROM jhnnsrs/tensorflow_conda:latest

# setup conda virtual environment
COPY ./requirements.yml /tmp/requirements.yml
RUN conda env create --name camera-seg -f /tmp/requirements.yml

RUN conda activate camera-seg
"""

BUILDERS = {
    "pip": {
        "gpu": None,
        "no-gpu": pip_no_gpu,
    },
    "conda": None,
    "poetry": {
        "gpu": None,
        "no-gpu": poetry_no_gpu,
    },
}


def build_dockerfile(
    manifest: Manifest, packager: Packager, gpu: bool, python_version: str
) -> str:
    """A function to build a dockerfile

    It delegates to the correct builder, based on the packager and gpu support


    Parameters
    ----------
    manifest : Manifest
        The manifest of the project
    packager : Packager
        The detected packager
    gpu : bool
        The detected gpu support
    python_version : str
        _description_

    Returns
    -------
    str
        The dockerfile

    Raises
    ------
    click.ClickException
        An exception if the packager is not supported
    """



    strategy = BUILDERS.get(packager, {})
    
    if not strategy:
        raise click.ClickException(f"Packager {packager} is not supported. Please create a issue on github and create your own Dockerfile for now.")
    
    builder = strategy.get(gpu and "gpu" or "no-gpu", None)

    if not builder:
        raise click.ClickException(
            f"Packager {packager} {gpu and 'with GPU Support'} is not supported. Please create a issue on github and create your own Dockerfile for now."
        )
    
    return builder(manifest, python_version)
   
        



def docker_file_wizard(manifest: Manifest, auto: bool = True) -> str:
    """A wizard to generate a dockerfile

    Parameters
    ----------
    manifest : Manifest
        The manifest of the project
    auto : bool, optional
        Should we autodefault prompts to true, by default True
    
    Returns
    -------
    str
        The dockerfile
    
    ------
    ImportError
        If toml is not installed
    """
    python_version = (
        f"{ sys.version_info.major}.{ sys.version_info.minor}.{ sys.version_info.micro}"
    )
    python_version = (
        python_version
        if auto
        else click.prompt(
            "Which python version to do you want to use?", default=python_version
        )
    )

    packager = None

    gpu = "gpu" in manifest.requirements

    # loading environment
    # check if running in conda
    conda_env = os.environ.get("CONDA_DEFAULT_ENV")
    if conda_env:
        click.echo(f"Found running in conda environment {conda_env}")

        packager = Packager.CONDA

    # Check if pyproject.toml exists
    if os.path.exists("pyproject.toml"):
        # Check if poetry is installed
        try:
            import toml
        except ImportError as e:
            raise ImportError("Please install toml to use this feature") from e

        # Check if poetry is used
        pyproject = toml.load("pyproject.toml")
        if "tool" in pyproject and "poetry" in pyproject["tool"]:
            click.echo("Found poetry project")
            do = True if auto else click.confirm("Do you want to use poetry?")
            if do:
                packager = Packager.POETRY

    if not packager:
        click.echo(
            "No advanced packager found. Considering using poetry "
            "with a pyproject.toml or conda with a environment.yml"
        )

        if os.path.exists("requirements.txt"):
            click.prompt("Found requirements.txt file in current directory. Using it.")

        packager = Packager.PIP


    if not click.confirm(
        f"Would you like to generate Template Dockerfile: Using {packager} with Python {python_version} {'and' if gpu else 'without'} GPU support"
    ):
        raise click.Abort("Aborted")
    
    dockfile = build_dockerfile(
        manifest, packager, gpu=gpu, python_version=python_version
    )


    return dockfile


def check_overwrite_dockerfile(ctx: Context,param: str, value: bool) -> bool:
    """A callback to check if the dockerfile should be overwritten

    Parameters
    ----------
    ctx : Context
        The context of the click command
    param : str
        The parameter name
    value : str
        The value of the parameter

    Returns
    -------
    bool
        Should we overwrite the dockerfile?
    """

    app_file = ctx.params["dockerfile"]
    if os.path.exists(app_file) and not value:
        should_overwrite = click.confirm(
            "Docker already exists. Do you want to overwrite?", abort=True
        )
        if should_overwrite:
            return True
        else:
            raise click.Abort()

    return value


@click.command()
@click.option("--dockerfile", help="The dockerfile to generate", default="Dockerfile")
@click.option(
    "--overwrite-dockerfile",
    "-o",
    help="Should we overwrite the existing Dockerfile?",
    is_flag=True,
    default=False,
    callback=check_overwrite_dockerfile,
)
@click.pass_context
def wizard(ctx: Context, dockerfile: str, overwrite_dockerfile: bool) -> None:
    """Runs the port wizard to generate a dockerfile to be used with port"""

    manifest = get_manifest(ctx)
    
    dockfile = docker_file_wizard(manifest)

    with open(dockerfile, "w") as file:
        file.write(dockfile)
